# -*- python -*-
# $RCSfile: color.py,v $
# $Revision: 1.25 $
# $Author: langer $
# $Date: 2004/12/07 21:06:06 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

import math, types
from common import registeredclass
from common.IO import parameter

FloatParameter = parameter.FloatParameter
FloatRangeParameter = parameter.FloatRangeParameter

# Color is a convertible registered class.  The same color can be
# parametrized as either rgb or hsv, or gray.

class Color(registeredclass.ConvertibleRegisteredClass):
    registry = []
    # Generic comparator for colors.  Subclasses must provide
    # getRed(), getGreen() and getBlue() functions, which they need
    # for other reasons anyways.
    def __cmp__(self,other):
        try:
            if self.getRed() < other.getRed(): return -1
            if self.getRed() > other.getRed(): return 1
            if self.getGreen() < other.getGreen(): return -1
            if self.getGreen() > other.getGreen(): return 1
            if self.getBlue() < other.getBlue(): return -1
            if self.getBlue() > other.getBlue(): return 1
        except AttributeError:
            return 1
        return 0
    # Also need to over-ride RegisteredClass's __eq__ and __ne__ functions.
    def __eq__(self,other):
        return self.__cmp__(other)==0
    def __ne__(self,other):
        return self.__cmp__(other)!=0
    
    def rgb(self):
        return (self.getRed(), self.getGreen(), self.getBlue())

Color.tip = "Various ways of describing a color."
Color.discussion = "<simpara>TODO: Add discussion.</simpara>"
        
# Structured type to hold the base values for color parameters,
# and which can be type-checked.  Also contains the comparator
# for colors.
class ColorValueBase:
    def __init__(self,red,green,blue):
        self.red = red
        self.green = green
        self.blue = blue
    def getRed(self):
        return self.red
    def getGreen(self):
        return self.green
    def getBlue(self):
        return self.blue
    # Comparator, same semantics as instance comparator, but usable
    # if you don't yet have an instance.  (Does this ever happen?)
    def __cmp__(self,other):
        try:
            if self.red < other.red: return -1
            if self.red > other.red: return 1
            if self.green < other.green: return -1
            if self.green > other.green: return 1
            if self.blue < other.blue: return -1
            if self.blue > other.blue: return 1
        except AttributeError:
            return 1                    # 'other' is (probably) not a CVB.
        return 0
    def __repr__(self):
        return "ColorValueBase(red=%f, green=%f, blue=%f)" % \
               (self.red, self.green, self.blue)
    
# Registration assigns the conversion routines, which make
# Color into a nontrivial convertible class.  "to_base" and
# "from_base" arguments are required at this level, in general.
# For the particular case of Color, however, all of the instances
# can answer "base" queries (r, g, or b) more efficiently at
# the instance level, so "to_base" functions are not provided.
class ColorRegistration(registeredclass.ConvertibleRegistration):
    def __init__(self, name, subclass, ordering,
                 from_base, params=[], tip=None, secret=0):
        registeredclass.ConvertibleRegistration.__init__(
            self, name, Color, subclass, ordering, from_base=from_base,
            params=params, tip=tip, secret=secret)

    def getParamValuesAsBase(self):
        temp = self()
        return ColorValueBase(temp.getRed(), temp.getGreen(), temp.getBlue())


class RGBColor(Color):
    """
    Initialize with key word arguments or rgb values:
      RGBColor(red, green, blue)
      RGBColor(red=x, green=y, blue=z)
    """
    def __init__(self, red, green, blue):
        self.red = red
        self.green = green
        self.blue = blue
    def getRed(self):
        return self.red
    def getGreen(self):
        return self.green
    def getBlue(self):
        return self.blue


# RGB conversion functions are nearly trivial.
def _rgb_from_base(rgb, base):
    if isinstance(base,ColorValueBase):
        rgb.setDefaultParams([base.red, base.green, base.blue])


ColorRegistration('RGBColor', RGBColor, 1,
                  params=[FloatRangeParameter('red', (0., 1., 0.01), 0.0,
                                              tip="Value of Red."),
                          FloatRangeParameter('green', (0., 1., 0.01), 0.0,
                                              tip="Value of Green."),
                          FloatRangeParameter('blue', (0., 1., 0.01), 0.0,
                                              tip="Value of Blue.")],
                  from_base=_rgb_from_base,
                  tip="Red, Green and Blue color.")



    
class Gray(Color):
    def __init__(self, value):
        self.value = value
    def getRed(self):
        return self.value
    def getGreen(self):
        return self.value
    def getBlue(self):
        return self.value

# Gray conversion functions.
def _gray_from_base(gray, base):
    if isinstance(base,ColorValueBase):
        g = (base.red+base.green+base.blue)/3.0
        gray.setDefaultParams([g])
        

ColorRegistration('Gray', Gray, 3,
                  params=[FloatRangeParameter('value', (0., 1., 0.01), 0.0,
                                              tip="Gray value.")],
                  from_base=_gray_from_base,
                  tip="Monochrome color.")

        

class HSVColor(Color):
    def __init__(self, hue, saturation, value):
        if hue >= 360:
            hue = hue % 360
        self.hue = hue
        self.saturation = saturation
        self.value = value
        self._rgb = None                # cache conversion to rgb

    def findrgb(self):
        ## HSV to RGB algorithm stolen from
        ## http://www.cs.rit.edu/~ncs/color/t_convert.html
        if self._rgb:
            return
        if self.saturation == 0.0: # gray
            self._rgb = (self.value, self.value, self.value)
        else:
            h = self.hue/60.
            i = int(math.floor(h))    # sextant of the color wheel
            f = h - i                 # fractional part of h
            p = self.value * (1 - self.saturation)
            q = self.value * (1 - self.saturation*f)
            t = self.value * (1 - self.saturation*(1-f))
            if (i == 0) or (i == 6):  # For h=360.0, can get i=6.
                self._rgb = (self.value, t, p)
            elif i == 1:
                self._rgb = (q, self.value, p)
            elif i == 2:
                self._rgb = (p, self.value, t)
            elif i == 3:
                self._rgb = (p, q, self.value)
            elif i == 4:
                self._rgb = (t, p, self.value)
            elif i == 5:
                self._rgb = (self.value, p, q)

    def getRed(self):
        self.findrgb()
        return self._rgb[0]
    def getGreen(self):
        self.findrgb()
        return self._rgb[1]
    def getBlue(self):
        self.findrgb()
        return self._rgb[2]



# HSV conversion functions are nontrivial.
def hsv_from_rgb(r, g, b):
     maxval = max((r,g,b))
     minval = min((r,g,b))
     #
     v = maxval
     #
     delta = maxval-minval
     #
     if maxval==0.0:
         (h,s,v) = (0.0, 0.0, 0.0) # If s=0, h and v don't matter.
     elif delta==0.0:  # It's gray, r=g=b.
         (h,s,v) = (0.0, 0.0, maxval)
     else:
         s = delta/maxval
         #
         if r==maxval:
             h = (g-b)/delta
         elif g==maxval:
             h = 2.0+(b-r)/delta
         else:
             h = 4.0+(r-g)/delta
         h *= 60.0
         if h<0:
             h+=360.0
     return (h, s, v)

def _hsv_from_base(hsv, base):
    if isinstance(base, ColorValueBase):
        (h, s, v) = hsv_from_rgb(base.red, base.green, base.blue)
        hsv.setDefaultParams([h,s,v])
    
ColorRegistration('HSVColor', HSVColor, 2,
                  params=[FloatRangeParameter('hue', (0.0, 360.0, 1.), 0.0,
                                              tip="Value of Hue."),
                          FloatRangeParameter('saturation', (0., 1., .01), 0.0,
                                              tip="Value of Saturation."),
                          FloatRangeParameter('value', (0., 1., .01),  0.0,
                                              tip="Value of ... hmmm ... Value.")],
                  from_base=_hsv_from_base,
                  tip="Hue, Saturation, and Value.")


#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#

# ColorParameter and NewColorParameter differ only in the widgets that
# are assigned to them in graphics mode.  The ColorParameter widget
# displays both the selected color and the old value.
# NewColorParameter only displays the selected color.

class ColorParameter(parameter.ConvertibleRegisteredParameter):
    def __init__(self, name, value=None, default=None, tip=None):
        parameter.ConvertibleRegisteredParameter.__init__(self, name, Color,
                                                          value, default, tip)
    def checker(self, x):
        if not isinstance(x, Color):
            parameter.raiseTypeError(type(x), "Color")
    def clone(self):
        return self.__class__(self.name, self.value, self.default, self.tip)

class NewColorParameter(ColorParameter): pass

#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#

### This is a test of secret registered classes.
##class Squant(Color): pass
##ColorRegistration('Squant', Squant, None, 1, secret=1)
                  

## Standard colors

black =   Gray(0.0)
white =   Gray(1.0)
red =     RGBColor(1.0, 0.0, 0.0)
green =   RGBColor(0.0, 1.0, 0.0)
blue =    RGBColor(0.0, 0.0, 1.0)
cyan =    RGBColor(0.0, 1.0, 1.0)
yellow =  RGBColor(1.0, 1.0, 0.0)
orange =  RGBColor(1.0, 0.5, 0.0)
magenta = RGBColor(1.0, 0.0, 1.0)
gray50 =  Gray(0.5)

from common import utils
utils.OOFdefine('black', black)
utils.OOFdefine('white', white)
utils.OOFdefine('red', red)
utils.OOFdefine('green', green)
utils.OOFdefine('blue', blue)
utils.OOFdefine('cyan', cyan)
utils.OOFdefine('yellow', yellow)
utils.OOFdefine('orange', orange)
utils.OOFdefine('magenta', magenta)
utils.OOFdefine('gray50', gray50)
