# -*- python -*-
# $RCSfile: materialListGUI.py,v $
# $Revision: 1.20 $
# $Author: reida $
# $Date: 2004/09/21 13:23:22 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from gtk import *

from common import debug
from common.IO import reporter
from common.guitop import top
from common.IO.mainmenu import OOF

# MaterialPropGUI is the GUI part of a property when it's
# part of a material.  Its gtk object is a GtkListItem object,
# and it hosts selection events, which are synchronized with
# the selection state of the materialmanager in /engine/materialmanager.py.
class MaterialPropGUI:
    def __init__(self, name, label=None, gui_props=None):
        debug.mainthreadTest()
        self.gtk = GtkListItem(label=label)
        self.name = name
        # At this point, we no longer have the propertyregistration object.
        # So, don't try to instance from here.
        # gui_props.add(self.gtk)
        self.label = label
        self.parent = None # Gets set by MaterialListGUI.add_prop.
        #
        # Connect local event handlers. 
        self.gtk.connect("button_press_event",self.on_buttonpress)
        # Callback ID's are stored here, but they're actually set
        # in the enclosing MaterialListGUI's "add_prop" method.
        self.select_CBID = None
        self.deselect_CBID = None 

    def on_buttonpress(self,gtk,event):
        pass
        # print "Buttonpress on material property."

    def __repr__(self):
        return "MaterialPropGUI('%s')" % self.name

# TODO LATER: The MaterialListGUI seems like a needlessly elaborate
# way of keeping track of the GUI selection state.  Much of the code
# probably dates back to the days when the selection state was shared
# with the main OOF menu.  These wrapper objects may be overkill.

# The MaterialListGUI is a list of MaterialPropGUI objects.
# The .gtk of this is the GtkMenuItem object added to the
# GtkOptionMenu on the materialsPage.
class MaterialListGUI:
    def __init__(self, parent, name, material, proplist):
        self.parent = parent
        self.name = name
        self.proplist = proplist 
        self.material = material # MaterialProps object.
        self.top = top()
        self.elements = {} # "elements" are local property GUI objects.
        self.current_element = None

        self.select_cb = None
        self.deselect_cb = None

        
    def __repr__(self):
        return "MaterialListGUI('%s')" % self.name

    # MaterialListGUI objects add and remove themselves from the
    # list, according to the selection state.
    def select(self):
        debug.mainthreadTest()
        for e in self.elements.values():
            self.proplist.add(e.gtk)
            e.gtk.show()

    def deselect(self):
        debug.mainthreadTest()
        for e in self.elements.values():
            # Suppress processing of "implicit deselection on removal".
            e.gtk.signal_handler_block(e.deselect_CBID)
            self.proplist.remove(e.gtk)
            e.gtk.signal_handler_unblock(e.deselect_CBID)


    # Set the selection state to be consistent with that of
    # the propertypane.  This is invoked just after material
    # selection or material creation.
    def selection_sync(self):
        current_prop = self.parent.parent.current_property()
        if current_prop:
            self.select_prop(current_prop[0])
        
        
    # In add_prop, selection is handled elsewhere, we
    # must set current_element to remain consistent.
    # "property" at this point is a MaterialPropGUI object.
    def add_prop(self, name, property):
        debug.mainthreadTest()
        self.elements[name]=property
        property.parent = self
        #
        property.select_CBID = property.gtk.connect("select",
                                                    self.GUI_select_prop,
                                                    name)
        property.deselect_CBID = property.gtk.connect("deselect",
                                                      self.GUI_deselect_prop,
                                                      name)
        
    def remove_prop(self, name):
        debug.mainthreadTest()
        self.deselect_prop(name)
        dead_prop = self.elements[name]
        dead_prop.gtk.disconnect(dead_prop.select_CBID)
        dead_prop.gtk.disconnect(dead_prop.deselect_CBID)
        dead_prop.gtk.destroy()
        del self.elements[name]
    
    # Property selection events have to be received here, because
    # this is where the property GUI objects live, but are processed
    # in the parent -- simply pass control to the parent.
    def GUI_select_prop(self, gtk, name):
        self.parent.prop_select(name)
        
    def GUI_deselect_prop(self, gtk, name):
        self.parent.prop_deselect(name)

    # Actual selection -- self.current_element is set here.
    def select_prop(self, propname):
        debug.mainthreadTest()
        try:
            # Can't use self.current_element because of implicit deselect.
            new_element = self.elements[propname]
        except KeyError:
            pass # Property is not present.  Do nothing, not an error.
        else:
            new_element.gtk.signal_handler_block(new_element.select_CBID)
            new_element.gtk.select()
            new_element.gtk.signal_handler_unblock(new_element.select_CBID)
            self.current_element = new_element

    def deselect_prop(self, propname):
        debug.mainthreadTest()
        try:
            element = self.elements[propname]
        except KeyError:
            pass # Deselected property not present, do nothing.
        else:
            # Only deselect if the property is present and current.
            if self.current_element == self.elements[propname]:
                self.current_element.gtk.signal_handler_block(
                    self.current_element.deselect_CBID)
                self.current_element.gtk.deselect()
                self.current_element.gtk.signal_handler_unblock(
                    self.current_element.deselect_CBID)
                self.current_element = None
                

    # Called when the "remove-selected" button is pressed.
    # Needs to be sure to remove the dictionary entry indicating
    # an instanced property type, also.
    def remove_current(self):
        debug.mainthreadTest()
        if self.current_element:
            # Pull the "gtk" out of current_element, because current_element
            # will be automatically reset to "None" (via "deselect") during
            # the destruction process.
            gtk = self.current_element.gtk
            element = self.current_element
            # Remove from property-tree-element's reference list.
            element.property.selector.remove_mat_entry()
            element.property = None # Break feedback loop.
            # Remove from local list.
            self.elements.remove(self.current_element)
            # Break circular reference chain?
            gtk.destroy()

        else:
            reporter.report("No property selected for removal.\n")

