# -*- python -*-
# $RCSfile: contourdisplay.py,v $
# $Revision: 1.72 $
# $Author: langer $
# $Date: 2004/12/07 21:06:07 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from SWIG.common import lock
from SWIG.engine.IO import contour
from common import color
from common import debug
from common import primitives
from common import registeredclass
from common import utils
from common.IO import colormap
from common.IO import display
from common.IO import output
from common.IO import parameter
from common.IO import reporter
from common.IO import xmlmenudump
from common.IO.automatic import *
from engine.IO import displaymethods
from engine.IO import outputClones
from types import *

Parameter = parameter.Parameter
IntParameter = parameter.IntParameter
IntRangeParameter = parameter.IntRangeParameter
FloatParameter = parameter.FloatParameter
AutoNumericParameter = parameter.AutoNumericParameter
ValueSetParameter = parameter.ValueSetParameter

contourparams = [
    output.ScalarOutputParameter('what', tip='Quantity to be plotted'),
    output.PositionOutputParameter('where', outputClones.posOutput,
                            tip="Plot at displaced or original position?"),
    AutoNumericParameter('min', automatic, 
                         tip="level of lowest contour, or 'automatic'"),
    AutoNumericParameter('max', automatic,
                         tip="level of highest contour, or 'automatic'"),
    ValueSetParameter('levels', 11,
                      tip="number of levels or list of levels (in [])"),
    IntParameter('nbins', 5, tip="number of internal subdivisions per element")
    ]

class ContourDisplay(displaymethods.MeshDisplayMethod):
    def __init__(self, what, where, min, max, levels, nbins):
        self.what = what
        self.where = where
        self.min = min
        self.max = max
        self.levels = levels
        self.nbins = nbins
        self.lock = lock.Lock()
        
        self.contourmaphidden = 0
        self.contourmap_user_hide = 0

        displaymethods.MeshDisplayMethod.__init__(self)

    def contour_capable(self, gfxwindow):
        if not self.incomputable(gfxwindow):
            return 1
        return 0
    
    def explicit_contour_hide(self):
        self.contourmap_user_hide = 1
        self.hide_contourmap()
        
    def hide_contourmap(self):
        self.contourmaphidden = 1

    def show_contourmap(self):
        self.contourmaphidden = 0

    def incomputable(self, gfxwindow):
        mesh = self.who().resolve(gfxwindow)
        return self.what.incomputable(mesh) or \
               displaymethods.MeshDisplayMethod.incomputable(self, gfxwindow)

    def find_levels(self, mesh, what):
        nlevels = self.levels
        clevels = None
##        if nlevels is automatic:
##            reporter.warn(
##                "levels=automatic is not yet implemented for contour plotting.")
##            # Not implemented yet.
##            # Should print an apology here.
##            return [], []
        if type(nlevels) == ListType or type(nlevels) == TupleType:
            clevels = nlevels
            nlevels = len(clevels)

        # Make a list of mastercoords of all nodes of all elements
        nodepoints = []
        for element in mesh.element_iterator():
            master = element.masterelement()
            el_mpos = []                # master coords of nodes in this element
            for n in range(master.nnodes()):
                el_mpos.append(master.get_protonode(n).mastercoord())
            nodepoints.append(el_mpos)

        # Evaluate the function at the nodes
        values = what.evaluate(mesh, mesh.element_iterator(), nodepoints)
        # Get function values grouped by element
        evalues = utils.unflatten(nodepoints, values)

##        vmax = max(values)
##        for (vals, element) in zip(evalues, mesh.element_iterator()):
##            if vmax in vals:
##                debug.fmsg('max found in element', element, vals)
##                break

        # Determine contour levels, if necessary
        if clevels == None:             # contours not specified by user
            # Find range of values.  This is just approximate, since the
            # interpolation within an element may give a value outside of
            # the range spanned by the nodal values.
            if self.min == automatic:
                vmin = float(min(values))
            else:
                vmin = float(self.min)
            if self.max == automatic:
                vmax = float(max(values))
            else:
                vmax = float(self.max)
            
            if nlevels == 1:
                clevels = [0.5*(vmax + vmin)]
            else:
                dz = (vmax - vmin)/(nlevels - 1)
                clevels = [vmin + i*dz for i in range(nlevels)]

        return clevels, evalues

##    def draw_subcells(self, mesh, device): # for debugging, presumably
##        for element in mesh.element_iterator():
##            cells = \
##                  element.masterelement().contourCells(self.nbins)
##            device.set_lineWidth(0)
##            device.set_lineColor(color.red)
##            for cell in cells:
##                cellpoints = self.where.evaluate(mesh, [element],
##                                                 [cell.corners])
##                device.draw_polygon(primitives.Polygon(cellpoints))
##            device.set_lineColor(color.black)


#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

class PlainContourDisplay(ContourDisplay):
    def __init__(self, what, where, min=automatic, max=automatic,
                 levels=automatic, nbins=5, width=0, color=color.black):
        ContourDisplay.__init__(self, what, where, min, max, levels, nbins)
        self.width = width
        self.color = color
        self.contour_max = None
        self.contour_min = None
        self.contour_levels = None
    def get_contourmap_info(self):
        return (self.contour_min, self.contour_max, self.contour_levels)
    def draw(self, gfxwindow, device):
        self.lock.acquire()
        try:
            mesh = self.who().resolve(gfxwindow).getObject()
            mesh.precompute()
    #        self.draw_subcells(mesh, device)
            device.comment("PlainContourDisplay")
            device.set_lineWidth(self.width)
            device.set_lineColor(self.color)

            # clevels is a list of contour values.
            # evalues is a list of lists of function values
            # for the nodes of each element.
            clevels, evalues = self.find_levels(mesh, self.what)
            for element in mesh.element_iterator():
                (contours, elmin, elmax)  = contour.findContours(
                    mesh, element, self.where,
                    self.what, clevels,
                    self.nbins, 0)
                for cntr in contours:
                    for loop in cntr.loops:
                        device.draw_polygon(loop)
                    for curve in cntr.curves:
                        device.draw_curve(curve)

            self.contour_min = min(clevels)
            self.contour_max = max(clevels)
            self.contour_levels = clevels
            contour.clearCache()
        finally:
            self.lock.release()

    def draw_contourmap(self, gfxwindow, device):
        self.lock.acquire()
        try:
            if self.contour_max is not None: # contours have been drawn
                aspect_ratio = gfxwindow.settings.aspectratio
                height = self.contour_max - self.contour_min
                width = height/aspect_ratio

                device.comment("Contourbar minimum: %g" % self.contour_min)
                device.comment("Contourbar maximum: %g" % self.contour_max)

                device.set_lineWidth(self.width)
                device.set_lineColor(self.color)

                for lvl in self.contour_levels:
                    r_lvl = lvl - self.contour_min
                    device.draw_curve( primitives.Curve(
                        [primitives.Point(0.0, r_lvl),
                         primitives.Point(width, r_lvl)] ))
        finally:
            self.lock.release()
            
        
registeredclass.Registration('ContourLine',
                             display.DisplayMethod,
                             PlainContourDisplay,
                             ordering=3.0,
                             layerordering=display.Linear(1),
                             whoclasses=('Mesh',),
                             params = contourparams +
                             [IntRangeParameter('width', (0,10), 0,
                                                tip="line width"),
                              color.ColorParameter('color', color.black,
                                                   tip="line color")],
                             tip="Draw contour lines for the given output data.",
                             discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/IO/plaincontour.xml')
                             )


#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

class FilledContourDisplay(ContourDisplay):
    def __init__(self, what, where, min=automatic, max=automatic,
                 levels=automatic, nbins=5, colormap=colormap.ThermalMap()):
        ContourDisplay.__init__(self, what, where, min, max, levels, nbins)
        self.colormap = colormap
        self.contour_max = None
        self.contour_min = None
        self.contour_levels = None
    # Return the colormap info generated by the previous "draw".
    def get_contourmap_info(self):
        return (self.contour_min, self.contour_max, self.contour_levels)
    def draw(self, gfxwindow, device):
        self.lock.acquire()
        try:
            self.contour_max = None
            self.contour_min = None
            self.contour_levels = []
            mesh = self.who().resolve(gfxwindow).getObject()
            mesh.precompute()
            device.comment("FilledContourDisplay")
            device.set_colormap(self.colormap)
            # clevels is a list of contour values.
            # evalues is a list of lists of function values
            # for the nodes of each element.
            clevels, evalues = self.find_levels(mesh, self.what)
            minval = min(clevels)
            maxval = max(clevels)
            valrange = maxval - minval
            if valrange == 0.0:
                valrange = 1
            factor = 1./valrange
            offset = -minval*factor

            for element, values in zip(mesh.element_iterator(), evalues):
                (contours, elmin, elmax)  = contour.findContours(
                    mesh, element, self.where,
                    self.what, clevels,
                    self.nbins, 1)

                # Before drawing anything, fill the element with the
                # largest contour value below its lowest detected value.
                vmin = min(values)
                prevcntour = None
                for cntour in contours:
                    if cntour.value > elmin:
                        if prevcntour:
                            device.set_fillColor(offset + prevcntour.value*factor)
                        else:
                            device.set_fillColor(0.0)
                        break
                    prevcntour = cntour
                # Find element perimeter
                edges = element.perimeter()
                mcorners = [[0.0]]*element.ncorners()
                corners = self.where.evaluate(mesh, edges, mcorners)
                device.comment("Element fill.")
                device.fill_polygon(primitives.pontify(primitives.Polygon(corners)))

                # Now fill contours
                for cntour in contours:
                    # This is harder than it looks.
                    if len(cntour.loops) == 1:
                        device.comment("Simple polygon.")
                        device.set_fillColor(offset + cntour.value*factor)
                        device.fill_polygon(cntour.loops[0])
                    elif len(cntour.loops) > 1:
                        device.comment("Multiple polygons.")
                        device.set_fillColor(offset + cntour.value*factor)
                        device.fill_polygon(cntour.loops) # Compound Polygon fill

            self.contour_min = minval
            self.contour_max = maxval
            self.contour_levels = clevels
            #  self.draw_subcells(mesh, device)
            contour.clearCache()
        finally:
            self.lock.release()

    def draw_contourmap(self, gfxwindow, device):
        # If the drawing failed, then contour_max won't have been set
        # yet.
        self.lock.acquire()
        try:
            if self.contour_max is not None:
                aspect_ratio = gfxwindow.settings.aspectratio
                height = self.contour_max - self.contour_min
                width = height/aspect_ratio

                device.comment("Colorbar minimum: %f" % self.contour_min)
                device.comment("Colorbar maximum: %f" % self.contour_max)
                device.set_colormap(self.colormap)

                for low, high in utils.list_pairs(self.contour_levels):
                    # Subtract "contour_min" off the y coords, so that
                    # the drawn object will include the point (0,0) --
                    # otherwise, the canvas bounds are wrong.
                    r_low = low-self.contour_min
                    r_high = high-self.contour_min

                    rect_bndy = map( lambda x: primitives.Point(x[0],x[1]),
                                     [ (0.0, r_low), (0.0, r_high),
                                       (width, r_high), (width, r_low) ] )

                    rectangle = primitives.Polygon(rect_bndy)
                    # In the collapsed case, height can be zero.  This is
                    # not hugely informative, but should be handled without
                    # crashing.
                    if height>0.0:
                        device.set_fillColor(r_low/height)
                    else:
                        device.set_fillColor(0.0)

                    device.fill_polygon(rectangle)
        finally:
            self.lock.release()

registeredclass.Registration('FilledContour',
                             display.DisplayMethod,
                             FilledContourDisplay,
                             ordering=3.1,
                             layerordering=display.Planar(1),
                             whoclasses = ('Mesh',),
                             params =
                             contourparams +
                             [Parameter('colormap', colormap.ColorMap,
                                        colormap.ThermalMap(),
                                        tip="color scheme")
                              ],
                             tip="Draw a filled contour plot of the given output data.",
                                                          discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/IO/filledcontour.xml')
                             )

