# -*- python -*-
# $RCSfile: skeletongroupmenu.py,v $
# $Revision: 1.24 $
# $Author: haan $
# $Date: 2004/12/09 22:52:00 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# This file builds the menu commands that operate the methods
# of the various selectable-groups, corresponding to methods
# of GenericGroupSet class in skeletongroups.py

from SWIG.common import ooferror
from common import debug
from common import utils
from common.IO import mainmenu
from common.IO import oofmenu
from common.IO import parameter
from common.IO import whoville
from common.IO import automatic
from common.IO import reporter
from engine import skeletongroups
from engine.IO import skeletongroupparams
import common.microstructure
import types

AutomaticNameParameter = parameter.AutomaticNameParameter
NodeGroupParameter = skeletongroupparams.NodeGroupParameter
SegmentGroupParameter = skeletongroupparams.SegmentGroupParameter
ElementGroupParameter = skeletongroupparams.ElementGroupParameter

nodegroupmenu = oofmenu.OOFMenuItem("NodeGroup", cli_only=1)
segmentgroupmenu = oofmenu.OOFMenuItem("SegmentGroup", cli_only=1)
elementgroupmenu = oofmenu.OOFMenuItem("ElementGroup", cli_only=1)

mainmenu.OOF.addItem(nodegroupmenu)
mainmenu.OOF.addItem(segmentgroupmenu)
mainmenu.OOF.addItem(elementgroupmenu)


########################################################################
# Generic callbacks usable by all classes of selectables.  Menu items
# must have a data item which is a string corresponding to the name
# of the selection set in the skeletoncontext, e.g. "nodesgroups",
# so that the menuitem callback can retrieve the right groupset.

# Make a new group name known.

def _new_group(menuitem, skeleton, name):
    skelc = whoville.getClass('Skeleton')[skeleton]
    groupset = getattr(skelc, menuitem.data)
    if groupset.isGroup(name):
        raise ooferror.ErrUserError("Group %s already exists." % group)
    else:
        groupset.addGroup(name)


# Remove a group.

def _remove_group(menuitem, skeleton, group):
    skelc = whoville.getClass('Skeleton')[skeleton]
    groupset = getattr(skelc, menuitem.data)
    if not groupset.isGroup(group):
        raise ooferror.ErrUserError("Group %s does not exist." % group)
    else:
        groupset.removeGroup(group)


# Copy an existing group.
def _copy_group(menuitem, skeleton, group, new_name):
    skelc = whoville.getClass('Skeleton')[skeleton]
    groupset = getattr(skelc, menuitem.data)
    if groupset.isGroup(new_name):
        raise ooferror.ErrUserError("Group %s already exists." % new_name)
    if groupset.isGroup(group):
        groupset.copyGroup(group, new_name)


# Add the current selection to the named group, creating
# the group if required.

def _add_selection_to_group(menuitem, skeleton, group):
    skelc = whoville.getClass('Skeleton')[skeleton]
    groupset = getattr(skelc, menuitem.data)
    if not groupset.isGroup(group):
        groupset.addGroup(group)
    groupset.addSelectionToGroup(group)

# Remove the currently-selected objects from the named group.
def _remove_selection_from_group(menuitem, skeleton, group):
    skelc = whoville.getClass('Skeleton')[skeleton]
    groupset = getattr(skelc, menuitem.data)
    # if not groupset.isGroup(group):
    #     groupset.removeGroup(group)
    groupset.removeSelectionFromGroup(group)

    
# Rename an existing group.
def _rename_group(menuitem, skeleton, group, new_name):
    skelc = whoville.getClass('Skeleton')[skeleton]
    groupset = getattr(skelc, menuitem.data)
    groupset.renameGroup(group, new_name)
##     if groupset.isGroup(new_name):
##         raise ooferror.ErrUserError("Group %s already exists." % new_name)
##     if groupset.isGroup(group):
##         groupset.renameGroup(group, new_name)
        

# Clear a group.

def _clear_group(menuitem, skeleton, group):
    skelc = whoville.getClass('Skeleton')[skeleton]
    groupset = getattr(skelc, menuitem.data)
    if not groupset.isGroup(group):
        raise ooferror.ErrUserError("Group %s does not exist." % group)
    groupset.clearGroup(group)


# Query a node group.
def _query_node_group(menuitem, skeleton, group):
    skelc = whoville.getClass('Skeleton')[skeleton]
    members = skelc.nodegroups.get_group(group)
    reporter.report(">>> ", len(members), " nodes")

def _query_sgmt_group(menuitem, skeleton, group):
    skelc = whoville.getClass('Skeleton')[skeleton]
    members = skelc.segmentgroups.get_group(group)
    reporter.report(">>> ", len(members), " segments")

def _query_elem_group(menuitem, skeleton, group):
    skelc = whoville.getClass('Skeleton')[skeleton]
    members = skelc.elementgroups.get_group(group)
    area = 0.0
    homog = 0.0
    for element in members:
        area += element.area()
        homog += element.homogeneity(skelc.getObject())
    if len(members):
        homog /= len(members)
    reporter.report(">>> ", len(members), " elements,", " area: ", area,
                    ", average homogeneity: ", homog)

#######################################

class GroupNameResolver:
    # Callable object to be used as the resolver for the
    # AutomaticNameParameter for group names.  Requires derived
    # classes that define getBasename().
    def __init__(self, attrname):
        # attrname is the name of the SkeletonContext attribute that
        # holds the groupset, eg, "nodegroups".
        self.attrname = attrname
    def __call__(self, param, startname):
        skelname = param.group['skeleton'].value
        skelcontext = whoville.getClass('Skeleton')[skelname]
        groupset = getattr(skelcontext, self.attrname)
        return utils.uniqueName(self.getBasename(param, startname),
                                groupset.allGroups())

class NewGroupNameResolver(GroupNameResolver):
    # Gets the default name from the constructor arguments.
    def __init__(self, basename, attrname):
        GroupNameResolver.__init__(self, attrname)
        # basename is the defaultname to be used for the groups.
        self.basename = basename
    def getBasename(self, param, startname):
        if param.automatic():
            return self.basename
        return startname

class CopyGroupNameResolver(GroupNameResolver):
    # Gets the default name from the object being copied (or renamed).
    def getBasename(self, param, startname):
        if param.automatic():
            return param.group['group'].value
        return startname
    
#######################################    
# Node-specific menu items:

new = oofmenu.OOFMenuItem("New_Group", cli_only=1,
                          callback=_new_group,
                          params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    AutomaticNameParameter("name", value=automatic.automatic,
                           resolver=NewGroupNameResolver("nodegroup",
                                                         "nodegroups"),
                           tip="Name of the group to be created.")),
                          help="Create a new node group.",
                          discussion="<para>A container for &nodes;. One can use &oof2; without using <link linkend='Section:Concepts:Skeleton:Groups'>groups</link> (it's not necessary to create them), but they certainly make some operations efficient, if they are to be operated on the same group of &skel; components.</para>")

new.data = "nodegroups"

nodegroupmenu.addItem(new)

rename = oofmenu.OOFMenuItem("Rename_Group", cli_only=1,
                             callback=_rename_group,
                             params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    NodeGroupParameter("group", tip="Name of the group to be renamed."),
    parameter.StringParameter("new_name", "",
                              tip="New name for the node group")),
                             help="Rename the group.",
                             discussion="<para>If you don't like the name, change it.</para>")
                             

rename.data = "nodegroups"

nodegroupmenu.addItem(rename)


copy = oofmenu.OOFMenuItem("Copy_Group", cli_only=1,
                           callback=_copy_group,
                           params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    NodeGroupParameter("group", tip="Name of the group to be copied."),
    AutomaticNameParameter("new_name",
                           resolver=CopyGroupNameResolver("nodegroups"),
                           value=automatic.automatic,
                           tip="Name for the copy.")),
                           help="Make a copy of the group.",
                           discussion="<para>If one wants to create a group of &nodes; and it will share some common &nodes; with the exisitng one, copying is better than creating it from the scratch, especillay the common &nodes; are hard to pick up.</para>")

copy.data = "nodegroups"

nodegroupmenu.addItem(copy)


remove = oofmenu.OOFMenuItem("Delete_Group", cli_only=1,
                             callback=_remove_group,
                             params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    NodeGroupParameter("group", tip="Name of the group to be deleted.")],
                             help="Remove the group.",
                             discussion="<para>If you don't like it all that much, why not just remove it.</para>")

remove.data = "nodegroups"

nodegroupmenu.addItem(remove)



addselect = oofmenu.OOFMenuItem("Add_to_Group", cli_only=1,
                                callback=_add_selection_to_group,
                                params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    NodeGroupParameter("group", tip="Name of the group.")],
                                help="Add selected nodes to the group.",
                                discussion="<para>It takes selected &nodes; and add them to the designated group.</para>") 

addselect.data = "nodegroups"

nodegroupmenu.addItem(addselect)


removeselect = oofmenu.OOFMenuItem("Remove_from_Group", cli_only=1,
                                callback=_remove_selection_from_group,
                                params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    NodeGroupParameter("group", tip="Name of the group")],
                                   help="Remove selected nodes from the group.",
                                   discussion="<para>To remove certain &nodes; from the group, select &nodes; to be removed and use this command. To display all the nodes that belong to the group, use <xref linkend='MenuItem.NodeSelection.Select_Group'/>.</para>") 

removeselect.data = "nodegroups"

nodegroupmenu.addItem(removeselect)


clear = oofmenu.OOFMenuItem("Clear_Group", cli_only=1,
                            callback=_clear_group,
                            params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    NodeGroupParameter("group", tip="Name of the group.")],
                            help="Clear the group.",
                            discussion="<para>This will clear the contents (&nodes;) not the group itself.</para>")

clear.data = "nodegroups"

nodegroupmenu.addItem(clear)

query = oofmenu.OOFMenuItem("Query_Group", cli_only=1,
                            callback=_query_node_group,
                            params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    NodeGroupParameter("group", tip="Name of the group.")],
                            help="Query the group.",
                            discussion="<para>For more useful information. Since &nodes; are fairly simple objects, all one needs to know should be found in the corresponding task page (<link linkend='Section:Tasks:SkeletonSelection'>Skeleton Selection</link>).</para>") 

query.data = "nodegroups"

nodegroupmenu.addItem(query)

#########################################################
# Segments.

new = oofmenu.OOFMenuItem("New_Group", cli_only=1,
                          callback=_new_group,
                          params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    AutomaticNameParameter("name",
                           value=automatic.automatic,
                           resolver=NewGroupNameResolver("segmentgroup",
                                                         "segmentgroups"),
                           tip="Name of the group.")),
                          help="Create a new segment group.",
                          discussion="<para>A container for &sgmts;.</para>")

new.data = "segmentgroups"

segmentgroupmenu.addItem(new)


rename = oofmenu.OOFMenuItem("Rename_Group", cli_only=1,
                             callback=_rename_group,
                             params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    SegmentGroupParameter("group", tip="Name of the group to be renamed."),
    parameter.StringParameter("new_name", "",
                              tip="New name for the segment group")),
                             help="Rename the group.",
                             discussion="<para>Renaming a &sgmt; group.</para>") 

rename.data = "segmentgroups"

segmentgroupmenu.addItem(rename)


copy = oofmenu.OOFMenuItem("Copy_Group", cli_only=1,
                           callback=_copy_group,
                           params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    SegmentGroupParameter("group", tip="Name of the group to be copied."),
    AutomaticNameParameter("new_name",
                           resolver=CopyGroupNameResolver("segmentgroups"),
                           value=automatic.automatic,
                           tip="New name for the copy.")),
                           help="Copy the group.",
                           discussion="<para>Copying a &sgmt; group.</para>") 

copy.data = "segmentgroups"

segmentgroupmenu.addItem(copy)


remove = oofmenu.OOFMenuItem("Delete_Group", cli_only=1,
                             callback=_remove_group,
                             params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    SegmentGroupParameter("group", tip="Name of the group.")],
                             help="Remove trhe group.",
                             discussion="<para>Deleting a &sgmt; group.</para>") 

remove.data = "segmentgroups"

segmentgroupmenu.addItem(remove)



addselect = oofmenu.OOFMenuItem("Add_to_Group", cli_only=1,
                                callback=_add_selection_to_group,
                                params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    SegmentGroupParameter("group", tip="Name of the group.")],
                                help="Add selected segments to the group.",
                                discussion="<para>Adding selected &sgmts; to a designated group. &sgmts; can be selected either from a <link linkend='Section:Tasks:SkeletonSelection'>task page</link> or from a <link linkend='Section:Graphics:SkeletonSelection'>toolbox</link>.</para>") 

addselect.data = "segmentgroups"

segmentgroupmenu.addItem(addselect)


removeselect = oofmenu.OOFMenuItem("Remove_from_Group", cli_only=1,
                                callback=_remove_selection_from_group,
                                params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    SegmentGroupParameter("group", tip="Name of the group.")],
                                   help="Remove selected segments from the group.",
                                   discussion="<para>Removing <link linkend='MenuItem:OOF.SegmentSelection'>selected</link> segments from the group.</para>")

removeselect.data = "segmentgroups"

segmentgroupmenu.addItem(removeselect)


clear = oofmenu.OOFMenuItem("Clear_Group", cli_only=1,
                            callback=_clear_group,
                            params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    SegmentGroupParameter("group", tip="Name of the group.")],
                            help="Clear the group.",
                            discussion="<para>Emptying the group.</para>") 

clear.data = "segmentgroups"

segmentgroupmenu.addItem(clear)


query = oofmenu.OOFMenuItem("Query_Group", cli_only=1,
                            callback=_query_sgmt_group,
                            params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    SegmentGroupParameter("group", tip="Name of the group.")],
                            help="Query the group.",
                            discussion="<para>For more information on the group.</para>") 

query.data = "segmentgroups"

segmentgroupmenu.addItem(query)

########################################################
# Elements

new = oofmenu.OOFMenuItem("New_Group", cli_only=1,
                          callback=_new_group,
                          params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    AutomaticNameParameter("name",
                           value=automatic.automatic,
                           resolver=NewGroupNameResolver("elementgroup",
                                                         "elementgroups"),
                           tip="Name for the element group to be created.")),
                          help="Create a new element group.",
                          discussion="<para>A container for &elems;.</para>") 

new.data = "elementgroups"


elementgroupmenu.addItem(new)


rename = oofmenu.OOFMenuItem("Rename_Group", cli_only=1,
                             callback=_rename_group,
                             params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    ElementGroupParameter("group", tip="Name of the group to be renamed."),
    parameter.StringParameter("new_name", "",
                              tip="New name for the element group")),
                             help="Rename the group.",
                             discussion="<para>If you don't like the name, rename it.</para>") 

rename.data = "elementgroups"

elementgroupmenu.addItem(rename)


copy = oofmenu.OOFMenuItem("Copy_Group", cli_only=1,
                           callback=_copy_group,
                           params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    ElementGroupParameter("group", tip="Name of the group to be copied."),
    AutomaticNameParameter("new_name",
                           resolver=CopyGroupNameResolver("elementgroups"),
                           value=automatic.automatic,
                           tip="Name for the copy.")),
                           help="Copy the group.",
                           discussion="<para>Making a duplicate of a group.</para>") 

copy.data = "elementgroups"

elementgroupmenu.addItem(copy)


remove = oofmenu.OOFMenuItem("Delete_Group", cli_only=1,
                             callback=_remove_group,
                             params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    ElementGroupParameter("group", tip="Name of the group.")],
                             help="Remove the group.",
                             discussion="<para>Removing an &elem; group.</para>") 

remove.data = "elementgroups"

elementgroupmenu.addItem(remove)



addselect = oofmenu.OOFMenuItem("Add_to_Group", cli_only=1,
                                callback=_add_selection_to_group,
                                params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    ElementGroupParameter("group", tip="Name of the group.")],
                                help="Add selected elements to the group.",
                                discussion="<para>Adding <link linkend='MenuItem:OOF.ElementSelection'>selected</link> elements to a group.</para>") 

addselect.data = "elementgroups"

elementgroupmenu.addItem(addselect)


removeselect = oofmenu.OOFMenuItem("Remove_from_Group", cli_only=1,
                                callback=_remove_selection_from_group,
                                params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    ElementGroupParameter("group", tip="Name of the group.")],
                                   help="Remove selected elements from the group.",
                                   discussion="<para>Removing <link linkend='MenuItem:OOF.ElementSelection'>selected</link> elements from a group..</para>") 

removeselect.data = "elementgroups"

elementgroupmenu.addItem(removeselect)


clear = oofmenu.OOFMenuItem("Clear_Group", cli_only=1,
                            callback=_clear_group,
                            params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    ElementGroupParameter("group", tip="Name of the group.")],
                            help="Clear the gorup.",
                            discussion="<para>Keep the group but not the &elems;.</para>")

clear.data = "elementgroups"

elementgroupmenu.addItem(clear)

query = oofmenu.OOFMenuItem("Query_Group", cli_only=1,
                            callback=_query_elem_group,
                            params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton'),
                          tip=parameter.emptyTipString),
    ElementGroupParameter("group", tip="Name of the group.")],
                            help="Query the group.",
                            discussion="<para>Useful informations such as total area, average <link linkend='Section:Concepts:Skeleton:Homogeneity'>homogeneity</link> of the group can be obtained.</para>") 

query.data = "elementgroups"

elementgroupmenu.addItem(query)
