# -*- python -*-
# $RCSfile: boundarybuilder.py,v $
# $Revision: 1.40 $
# $Author: langer $
# $Date: 2004/11/12 22:16:30 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# This file defines registered class objects which actually
# build boundaries.   Structure is similar to skeletionselectionmod.

from SWIG.common import ooferror
from SWIG.common import switchboard
from common import director
from common import registeredclass
from common.IO import automatic
from common.IO import parameter
from common.IO import selectionplaceholder
from common.IO import whoville
from engine import skeletoncontext
from engine import skeletonsegment # for the segSequence function.
from engine.IO import skeletongroupparams
import types

########################
# Boundary naming components.

# This resolver can return "None" in the skelname==None case, which
# occurs when the registeredclassfactory switches around.
def bdynameresolver(param, startname):
    if param.automatic():
        basename='boundary'
    else:
        basename = startname
    skelname = param.group['skeleton'].value
    skelcontext = skeletoncontext.skeletonContexts[skelname]
    return skelcontext.uniqueBoundaryName(basename)


skeletonparam = whoville.WhoParameter(
    'skeleton', skeletoncontext.skeletonContexts,
    tip="Specify the skeleton where the boundary resides.")

nameparam = parameter.AutomaticNameParameter('name', bdynameresolver,
                                             value=automatic.automatic,
                                             tip="Name for this boundary.")

paramgroup = parameter.ParameterGroup(skeletonparam, nameparam)

########################


class BoundaryConstructor(registeredclass.RegisteredClass):
    registry = []

BoundaryConstructor.tip = "Tools to build Boundary (a specified region in Skeleton to which Boundary Conditions are applied)."
BoundaryConstructor.discussion = "<para>TODO: Add discussion</para>"

# Utility function -- given a bunch of segments and an orientation,
# it returns (startnode, seg_list), where seg_list is suitably
# ordered, and startnode is at the start.  Throws an exception if
# it fails.
def _segset2seglist(seg_set, direction):

    if len(seg_set)==0:
        raise ooferror.ErrUserError(
            "Attempt to sequence null segment set.")
    
    (seg_list, node_list) = skeletonsegment.segSequence(seg_set)
    
    # At this point, seg_list is an ordered list of adjacent
    # segments, and "node_list" is the corresponding set of nodes.
    # The path is a loop if node_list[0]==node_list[-1], otherwise
    # it's a simple path.

    # We may want to reverse this list, depending on the
    # setting of "direction".
    firstnode = node_list[0]
    lastnode = node_list[-1]
    
    startnode = firstnode
    if firstnode != lastnode:
        if direction == director.Director('Left to right'):
            if firstnode.position().x > lastnode.position().x:
                seg_list.reverse()
                startnode = lastnode
        elif direction == director.Director('Right to left'):
            if firstnode.position().x < lastnode.position().x:
                seg_list.reverse()
                startnode = lastnode
        elif direction == director.Director('Top to bottom'):
            if firstnode.position().y < lastnode.position().y:
                seg_list.reverse()
                startnode = lastnode
        elif direction == director.Director('Bottom to top'):
            if firstnode.position().y > lastnode.position().y:
                seg_list.reverse()
                startnode = lastnode
        else:
            # User specified clockwise or counterclockwise for a
            # non-loop segment set.  This is an error.
            raise ooferror.ErrUserError(
                "Clockwise or counterclockwise is for closed loops.")

    else: # firstnode==lastnode, loop case.
        startnode = None
        area = 0.0
        p0 = node_list[0].position()
        for i in range(1,len(node_list)):
            p1 = node_list[i].position()
            area += p0.x * p1.y - p1.x * p0.y
            p0 = p1
            
        if direction == director.Director('Clockwise'):
            if area > 0.0:
                seg_list.reverse()
        elif direction == director.Director('Counterclockwise'):
            if area < 0.0:
                seg_list.reverse()
        else:
            # User specified an endpoint orientation on a loop.
            raise ooferror.ErrUserError(
                "Closed loops need clockwise or counterclockwise direction.")

    return (startnode, seg_list)


# ## ### #### ##### ###### ####### ######## ####### ###### ##### #### ### ## #


# Convert a group name and skeleton context into an actual list of segments.
def segments_from_seg_aggregate(skelcontext, group):
  if group == selectionplaceholder.selectionPlaceHolder:
      seg_set = skelcontext.segmentselection.retrieve()
  else:
      seg_set = skelcontext.segmentgroups.get_group(group)
  return seg_set

# Build an edge boundary from a segment set.
class EdgeFromSegments(BoundaryConstructor):
    def __init__(self, group, direction):
        self.group = group
        self.direction = direction

    def __call__(self, skelcontext, name):
        skelobj = skelcontext.getObject()

        seg_set = segments_from_seg_aggregate(skelcontext, group)
        
        (startnode, seg_list) = _segset2seglist(seg_set, self.direction)
        # At this point, we have a correctly-sequenced list of segments.
        # Actually create the boundary.  The context will create it in
        # the underlying object.
        skelcontext.createEdgeBoundary(name, seg_list, startnode)
    


registeredclass.Registration("Edge boundary from segments",
                             BoundaryConstructor,
                             EdgeFromSegments,
                             ordering=100,
                             params = [
    skeletongroupparams.SegmentAggregateParameter('group', tip="Construct the boundary from these segments"),
    director.DirectorParameter('direction', director.Director,
                               director.Director('Clockwise'),
                               tip="Direction of Boundary.")],
                             tip="Construct an edge boundary from a set of segments.")


    
# ## ### #### ##### ###### ####### ######## ####### ##### #### ### ## #

  
# Finds the perimeter of an elementgroup.  Requires that the element
# group have a topologically simple (non-self-intersecting,
# non-disjoint) boundary.

# Utility function for extracting a segment list from elements.
# Also used by the DirectorWidget.
def segments_from_el_aggregate(skelcontext, group):
    seg_set = {}

    
    if group == selectionplaceholder.selectionPlaceHolder:
        elements = skelcontext.elementselection.retrieve()
    else:
        elements = skelcontext.elementgroups.get_group(group)
        
    # We want to retain the boundary segments -- these occur
    # exactly once in the element set, and all the others occur
    # exactly twice, so iterate over all the segments, adding them
    # to a dictionary, unless they already appear, in which case
    # remove them.
    for e in elements:
        for s in e.getSegments(skelcontext.getObject()):
            try:
                del seg_set[s]
            except KeyError:
                seg_set[s]=1

    return seg_set.keys()
                
class EdgeFromElements(BoundaryConstructor):
    def __init__(self, group, direction):
        self.group = group
        self.direction = direction

    def __call__(self, skelcontext, name):
        seg_list = segments_from_el_aggregate(skelcontext, self.group)

        (startnode, seg_list) = _segset2seglist(seg_list, self.direction)

        skelcontext.createEdgeBoundary(name, seg_list, startnode)


registeredclass.Registration("Edge boundary from elements",
                             BoundaryConstructor,
                             EdgeFromElements,
                             ordering=101,
                             params = [
    skeletongroupparams.ElementAggregateParameter('group', tip="Construct the boundary from these elements"),
    director.DirectorParameter('direction', director.Director,
                               director.Director('Clockwise'),
                               tip="Direction of Boundary.")],
                             tip="Construct an edge boundary surrounding a set of elements.")


# ## ### #### ##### ###### ####### ######## ####### ###### ##### #### ### ## #

def segments_from_node_aggregate(skelcontext, group):
    seg_set = {}
    if group == selectionplaceholder.selectionPlaceHolder:
        nodes = skelcontext.nodeselection.retrieve()
    else:
        nodes = skelcontext.nodegroups.get_group(self.group)
        
    for n in nodes:
        for s in n.localSegments(skelcontext.getObject()):
            try:
                seg_set[s] += 1
            except KeyError:
                seg_set[s] = 1

    # Each segment which occurs twice is one we want, because
    # both ends of it are in the node set.
    good_segs = [ s for s in seg_set.keys() if seg_set[s]==2]

    return good_segs


class EdgeFromNodes(BoundaryConstructor):
    def __init__(self, group, direction):
        self.group = group
        self.direction = direction

    def __call__(self, skelcontext, name):
        skelobj = skelcontext.getObject()

        seg_set = segments_from_node_aggregate(skelcontext, self.group)
        
        (startnode, seg_list) = _segset2seglist(seg_set, self.direction)

        skelcontext.createEdgeBoundary(name, seg_list, startnode)
        

registeredclass.Registration("Edge boundary from nodes",
                             BoundaryConstructor,
                             EdgeFromNodes,
                             ordering=102,
                             params=[
    skeletongroupparams.NodeAggregateParameter('group', tip="Node group from which to deduce segments"),
    director.DirectorParameter('direction', director.Director,
                               director.Director('Clockwise'),
                               tip="Direction of Boundary.")],
                             tip="Construct an edge boundary from a collection of nodes.")




# ## ### #### ##### ###### ####### ######## ####### ###### ##### #### ### ## #


# Point boundary construction can be done from either
# segments or nodes, both cases are very simple.
class PointFromNodes(BoundaryConstructor):
    def __init__(self, group):
        self.group = group

    def __call__(self, skelcontext, name):
        skelobj = skelcontext.getObject()

        if self.group == selectionplaceholder.selectionPlaceHolder:
            nodelist = skelcontext.nodeselection.retrieve()
        else:
            nodelist = skelcontext.nodegroups.get_group(self.group)

        skelcontext.createPointBoundary(name, nodelist)

registeredclass.Registration("Point boundary from nodes",
                             BoundaryConstructor,
                             PointFromNodes,
                             ordering=103,
                             params = [
    skeletongroupparams.NodeAggregateParameter('group', tip="Construct the boundary from these nodes")],
                             tip="Construct a point boundary from a set of nodes.")


class PointFromSegments(BoundaryConstructor):
    def __init__(self, group):
        self.group = group

    def __call__(self, skelcontext, name):
        skelobj = skelcontext.getObject()

        nodes = {}

        if self.group == selectionplaceholder.selectionPlaceHolder:
            segments = skelcontext.segmentselection.retrieve()
        else:
            segments = skelcontext.segmentgroups.get_group(self.group)

        for s in segments:
            nodes[s.nodes()[0]]=s
            nodes[s.nodes()[1]]=s
                
        skelcontext.createPointBoundary(name, nodes.keys() )

registeredclass.Registration("Point boundary from segments",
                             BoundaryConstructor,
                             PointFromSegments,
                             ordering=104,
                             params = [
    skeletongroupparams.SegmentAggregateParameter('group', tip="Construct the boundary from these segments")],
                             tip="Construct a point boundary from the endpoints of a set of segments.",
                             boundarytype="edge")

class PointFromElements(BoundaryConstructor):
    def __init__(self, group):
        self.group = group

    def __call__(self, skelcontext, name):
        skelobj = skelcontext.getObject()

        if self.group == selectionplaceholder.selectionPlaceHolder:
            elements = skelcontext.elementselection.retrieve()
        else:
            elements = skelcontext.elementgroups.get_group(self.group)

        nodes = {}
        for element in elements:
            for i in range(element.nnodes()):
                n0 = element.nodes[i]
                n1 = element.nodes[(i+1)%element.nnodes()]
                seg = skelobj.getSegment(n0, n1)
                # A segment is on the boundary of the selection if it
                # belongs to only one element.
                n = 0
                for el in seg.getElements():
                    if el in elements: n += 1
                if n == 1:
                    nodes[n0] = None
                    nodes[n1] = None

        skelcontext.createPointBoundary(name, nodes.keys())

registeredclass.Registration("Point boundary from elements",
                             BoundaryConstructor,
                             PointFromElements,
                             ordering=105,
                             params = [
        skeletongroupparams.ElementAggregateParameter('group', tip="Construct the boundary from these elements")],
                             tip="Construct a point boundary surrounding a set of elements.")
