// -*- C++ -*-
// $RCSfile: property.C,v $
// $Revision: 1.30 $
// $Author: reida $
// $Date: 2004/10/18 14:26:41 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#include "common/IO/threadstate.h"
#include "common/swiglib.h"
#include "common/trace.h"
#include "engine/femesh.h"
#include "engine/field.h"
#include "engine/material.h"
#include "engine/property.h"
#include "engine/ooferror.h"
#include "engine/IO/propertyoutput.h"
using namespace std;

// It appears the PyObject * can't be const, because 
// PyObject_GetAttrString doesn't take a const argument.
Property::Property(const string &nm, PyObject *registration)
  : name_(nm), registration_(registration), active_(false),
    fields_reqd(0)
{ 
  acquirePyLock();
  // registry.classobj.__name__, viewed through C++...
  classname_ = PyString_AsString(
			 PyObject_GetAttrString(
				PyObject_GetAttrString(registration,"subclass"),
				"__name__"));
  // registry.modulename    
  modulename_ = PyString_AsString(
			  PyObject_GetAttrString(registration,"modulename"));

  Py_INCREF(registration_);
  releasePyLock();
}

// Note: There appears to be some kind of bug somewhere involving
// property registration reference counts -- if the incref/decref is
// left out of this code, then at apparently-random points, all the
// property *registration* entries associated with properties placed
// in materials via "newMaterial" calls will lose all their data
// members.  The incref/decref here is certainly correct, but probably
// not the real answer, since several other references to registration
// entries exist (e.g. in the PropertyRegistration object's ".data"
// LabelTree, and the PropertyPtr class's "registry" list), so the
// reference count should always be at least 2 even in the absence of
// an incref in the Property class.  
//
//   The observed behavior might be caused by the PropertyRegistration's
// (Python) destructor getting called too early somehow, demolishing
// the attributes but not removing the reference, if for instance
// there were some kind of bug in the Python 2.1.2 and 2.1.3
// interpreters.  Additional circumstantial evidence for this is that
// the bug has never occurred in Python 2.2.
Property::~Property() 
{
  acquirePyLock();
  Py_DECREF(registration_);
  releasePyLock();
}


// Additional note on the mystery bug mentioned above: It can be
// induced by making exactly two calls to this routine from the
// registration retrieval block in MaterialManager.new_material.  One,
// three, or four calls are OK -- that's as many as I tried.
PyObject *Property::registration() const {
  acquirePyLock();
  Py_INCREF(registration_);
  releasePyLock();
  return registration_;
}

void Property::require_field(const Field &field) {
  for(int i=0; i<fields_reqd.size(); i++)
    if(*fields_reqd[i] == field)
      return;			// don't list a field more than once
  fields_reqd.push_back(&field);
}

// A property is active if all fields that it uses are defined.

bool Property::is_active(const FEMesh *mesh) const {
  //Trace("Property::active, name is " + name_); 
  for(int i=0; i<fields_reqd.size(); i++) {
    if(!mesh->is_defined_field(*fields_reqd[i])) {
//       std::cerr << "Property::is_active: " << *fields_reqd[i] << " is not defined" << std::endl;
      return false;
    }
  }
  return true;
}

void Property::find_active(const FEMesh *mesh) {
  active_ = is_active(mesh);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void Property::bookkeeping(Material *material) {
  // Read data from the Property's Registration object (python) and do
  // all the bookkeeping and cross-referencing required by the
  // Material.

  acquirePyLock();

  PyObject *ptype = PyObject_CallMethod(registration(), "propertyType", 0);
  char *typestring = PyString_AsString(ptype);
  material->registerPropertyType(this, typestring);
  Py_XDECREF(ptype);
  

  // Read the list of required fields, and call
  // Field::registerProperty() for each of them.

  // First, call PropertyRegistration.fields(), which returns a list
  // of swigged fields.
  PyObject *list = PyObject_CallMethod(registration(), "fields", 0);

  // Loop over the list.
  int nfields = PyList_GET_SIZE(list);
  for(int i=0; i<nfields; i++) {
    // Extract the swigged ptr object.
    PyObject *swigobj = PyObject_GetAttrString(PyList_GET_ITEM(list,i),"this");
    char *swigstr = PyString_AsString(swigobj);
    Py_XDECREF(swigobj);
    // Convert it to the actual C++ Field object.
    Field *field;
    SWIG_GetPtr(swigstr, (void**)&field, "_Field_p");
    field->registerProperty(this);
  }
  Py_XDECREF(list);

  // Do the same thing for the fluxes that this Property contributes to.
  list = PyObject_CallMethod(registration(), "fluxes", 0); // list of Fluxes
  int nfluxes = PyList_GET_SIZE(list);
  for(int i=0; i<nfluxes; i++) {
    PyObject *swigobj = PyObject_GetAttrString(PyList_GET_ITEM(list,i),"this");
    char *swigstr = PyString_AsString(swigobj);
    Py_XDECREF(swigobj);
    Flux *flux;
    SWIG_GetPtr(swigstr, (void**)&flux, "_Flux_p");
    material->registerFlux(this, flux);
  }
  Py_XDECREF(list);

  // And for the PropertyOutputs, which are listed by name in the
  // PropertyRegistration.
  list = PyObject_CallMethod(registration(), "outputs", 0);
  int noutputs = PyList_GET_SIZE(list);
  for(int i=0; i<noutputs; i++) {
    material->registerOutput(this, PyString_AsString(PyList_GET_ITEM(list,i)));
  }
  Py_XDECREF(list);
  
  releasePyLock();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

std::ostream &operator<<(std::ostream &os, const Property &prop) {
  return os << "Property(" << prop.name() << ")";
}
