# -*- python -*-
# $RCSfile: skeletonselectionmod.py,v $
# $Revision: 1.47 $
# $Author: haan $
# $Date: 2004/11/19 22:52:33 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from SWIG.common import switchboard
from common import debug
from common import enum
from common import primitives
from common import registeredclass
from common import utils
from common.IO import parameter
from common.IO import whoville
from common.IO import xmlmenudump
from engine import skeletonelement
from engine.IO import skeletongroupparams
import types, random

# TODO LATER: Can the Segment/Element/Node operations here all be derived
# from common SkeletonSelectable classes?

# The way this is going to work is that there will be One True Menu
# for nodeselectionmodifiers, segmentselectionmodifers, and
# elementselectionmodifiers, and each of these registered classes will
# add itself to the menu, taking as an argument the relevant
# skeletoncontext.  It will then operate on the skeletoncontext in the
# usual way.

# This is the function that actually runs a selection modification.
# It is the menu callback for the automatically-generated menu
# items in engine/IO/skeletonselectmenu.py.  This one routine
# works for node, segment, and element selections.

def modify(menuitem, skeleton, **params):
    registration = menuitem.data
    modifier = registration(**params)
    skelcontext = whoville.getClass('Skeleton')[skeleton]
    selection = modifier.getSelection(skelcontext)
    selection.begin_writing()
    try:
        modifier(skelcontext, selection)
    finally:
        selection.end_writing()

    selection.mode().modifierApplied(modifier) # sends switchboard signal
    selection.signal()

###########################################

class NodeSelectionModifier(registeredclass.RegisteredClass):
    registry = []
    def getSelection(self, skelcontext):
        return skelcontext.nodeselection

class NodeFromSelectedSegments(NodeSelectionModifier):
    def __call__(self, skeleton, selection):
        nodes = {}
        for segment in skeleton.segmentselection.retrieve():
            nodes[segment.nodes()[0]] = None
            nodes[segment.nodes()[1]] = None
        selection.start()
        selection.clear()
        selection.select(nodes.keys())

registeredclass.Registration('Select From Selected Segments',
                             NodeSelectionModifier,
                             NodeFromSelectedSegments, ordering=0.1,
                             tip="Select nodes from selected segments.",
                             discussion="<para>All the associated nodes of selected segments will be selected.</para>")


class NodeFromSelectedElements(NodeSelectionModifier):
    def __init__(self, internal=0, boundary=1):
        self.internal = internal
        self.boundary = boundary

    def getAllNodes(self, context):
        nodes = {}
        for element in context.elementselection.retrieve():
            for nd in element.nodes:
                nodes[nd] = None
        return nodes.keys()

    def getBoundaryNodes(self, context):
        bound = {}
        skel = context.getObject()
        for element in context.elementselection.retrieve():
            for i in range(element.nnodes()):
                n0 = element.nodes[i]
                n1 = element.nodes[(i+1)%element.nnodes()]
                seg = skel.getSegment(n0, n1)
                # A segment is on the boundary of the selection if it
                # belongs to only one element.
                n = 0
                for el in seg.getElements():
                    if el.selected: n += 1
                if n == 1:
                    bound[n0] = None
                    bound[n1] = None
        return bound.keys()

    def getInternalNodes(self, context, allnodes):
        bound = self.getBoundaryNodes(context)
        internal = []
        for nd in allnodes:
            if nd not in bound:
                internal.append(nd)
        return internal

    def __call__(self, skeleton, selection):
        if self.internal and self.boundary:
            selected = self.getAllNodes(skeleton)
        elif not self.internal and self.boundary:
            selected = self.getBoundaryNodes(skeleton)
        elif self.internal and not self.boundary:
            allnodes = self.getAllNodes(skeleton)
            selected = self.getInternalNodes(skeleton, allnodes)
        else:
            selected = []
        selection.start()
        selection.clear()
        selection.select(selected)

registeredclass.Registration('Select From Selected Elements',
                             NodeSelectionModifier,
                             NodeFromSelectedElements, ordering=1,
                             params = [
    parameter.BooleanParameter('internal', 0, tip='Select internal nodes.'),
    parameter.BooleanParameter('boundary', 1, tip='Select boundary nodes.')],
                             tip="Select nodes from selected elements.",
                             discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu-nodes_from_elements.xml'))


class SelectInternalBoundaryNodes(NodeSelectionModifier):
    def __call__(self, skeleton, selection):
        skel = skeleton.getObject()
        nodelist = []
        for node in skel.nodes:
            elements = node.neighborElements()
            cat = elements[0].dominantPixel(skel)
            for element in elements[1:]:
                if cat != element.dominantPixel(skel):
                    nodelist.append(node)
                    break
        selection.start()
        selection.clear()
        debug.fmsg(selection)
        selection.select(nodelist)
                    
registeredclass.Registration(
    'Select Internal Boundaries',
    NodeSelectionModifier, SelectInternalBoundaryNodes,
    ordering=2.5,
    tip="Select all nodes on material or group boundaries.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu-boundary_nodes.xml'))

class ExpandCriterion(registeredclass.RegisteredClass):
    registry = []
    def expand(self, skeleton):
        pass

ExpandCriterion.tip = "Criteria for expanding node selection (see <xref linkend='MenuItem:OOF.NodeSelection.Expand_Node_Selection'/>)."
ExpandCriterion.discussion = "<para>See <link linkend='RegisteredClass:ExpandByElements'><classname>ExpandByElements</classname></link> and <link linkend='RegisteredClass:ExpandBySegments'><classname>ExpandBySegments</classname></link></para>"

class ExpandByElements(ExpandCriterion):
    # Boundary nodes : when not all of their neighbor elements are
    # selected.
    def boundary(self, node):
        for n in node.neighborElements():
            for nd in n.nodes:
                if not nd.selected:
                    return 1
        return 0
    
    def expand(self, skeleton):
        nodes = {}
        # Get the boundary nodes
        bnodes = [] 
        for node in skeleton.nodeselection.retrieve():
            if self.boundary(node):
                bnodes.append(node)
            else:
                nodes[node] = None
        for bn in bnodes:
            for ne in bn.neighborElements():
                for nd in ne.nodes:
                    nodes[nd] = None
        return nodes.keys()

registeredclass.Registration(
    'By Elements',
    ExpandCriterion, ExpandByElements,
    ordering=0,
    tip="Expand node selection by selecting all nodes of neighboring elements.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg-expand_by_elements.xml'))

class ExpandBySegments(ExpandCriterion):
    # Boundary nodes : when not all of their neighbor nodes are
    # selected.
    def boundary(self, node, skeleton):
        for n in node.neighborNodes(skeleton):
            if not n.selected:
                return 1
        return 0
    
    def expand(self, skeleton):
        nodes = {}
        # Get the boundary nodes
        bnodes = [] 
        for node in skeleton.nodeselection.retrieve():
            if self.boundary(node, skeleton.getObject()):
                bnodes.append(node)
            else:
                nodes[node] = None
        for bn in bnodes:
            for n in bn.neighborNodes(skeleton.getObject()):
                nodes[n] = None
            nodes[bn] = None
        return nodes.keys()

registeredclass.Registration(
    'By Segments',
    ExpandCriterion, ExpandBySegments,
    ordering=1,
    tip="Expand node selection by selecting all nodes of neighboring segments.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg-expand_by_segments.xml'))
    
class ExpandNodeSelection(NodeSelectionModifier):
    def __init__(self, criterion):
        self.criterion = criterion
    def __call__(self, skeleton, selection):
        selected = self.criterion.expand(skeleton)
        selection.start()
        selection.clear()
        selection.select(selected)        

registeredclass.Registration(
    'Expand Node Selection',
    NodeSelectionModifier, ExpandNodeSelection,
    ordering=2.6,
    params=[parameter.Parameter("criterion", ExpandCriterion)],
    tip="Exapnd node selection by either neighboring elements or segments.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu-expand_node.xml'))

# Select the indicated group.
class NodeSelectGroup(NodeSelectionModifier):
    def __init__(self, group):
        self.group = group
    def __call__(self, skeleton, selection):
        # Retrieve the members first -- if an exception occurs, the
        # system state will be as before.
        members = skeleton.nodegroups.get_group(self.group)
        selection.start()
        selection.clear()
        selection.select(members)

registeredclass.Registration('Select Group',
                             NodeSelectionModifier, NodeSelectGroup,
                             ordering=3,
                             params=[
    skeletongroupparams.NodeGroupParameter('group', tip="Name of the group.")],
                             tip='Select the members of a group, discarding the current selection.',
                             discussion="<para>Select nodes from a selected <link linkend='Section:Concepts:Skeleton:Groups'>group</link> <emphasis>BUT</emphasis> not before the current selection is wiped out -- you've been warned.</para>")


# Unselect the indicated group.
class NodeDeselectGroup(NodeSelectionModifier):
    def __init__(self, group):
        self.group = group
    def __call__(self, skeleton, selection):
        members = skeleton.nodegroups.get_group(self.group)
        selection.start()
        selection.deselect(members)

registeredclass.Registration('Unselect Group',
                             NodeSelectionModifier, NodeDeselectGroup,
                             ordering=4,
                             params=[
    skeletongroupparams.NodeGroupParameter('group', tip="Name of the group.")],
                             tip='Unselect the members of a group.',
                             discussion="<para>Deselect any nodes that are selected currently and at the same time, are the members of the specified <link linkend='Section:Concepts:Skeleton:Groups'>group</link>.</para>")


# Add the group to the selection, retaining the current selection.
class NodeAddSelectGroup(NodeSelectionModifier):
    def __init__(self, group):
        self.group = group
    def __call__(self, skeleton, selection):
        members = skeletion.nodegroups.get_group(self.group)
        selection.start()
        # Minor inefficiency, this reselects already-selected group members.
        selection.select(members)


registeredclass.Registration('Add Group',
                             NodeSelectionModifier, NodeAddSelectGroup,
                             ordering=5,
                             params=[
    skeletongroupparams.NodeGroupParameter('group', tip="Name of the group.")],
                             tip='Select the members of a group, retaining the current selection.',
                             discussion="<para>Unlike <xref linkend='MenuItem:OOF.NodeSelection.Select_Group'/>, it just add nods from the specified <link linkend='Section:Concepts:Skeleton:Groups'>group</link> to the current selection.</para>")


# Select the intersection of the group and the selection.
class NodeIntersectGroup(NodeSelectionModifier):
    def __init__(self, group):
        self.group = group
    def __call__(self, skeleton, selection):
        nlist = skeleton.nodegroups.get_group(self.group)
        ilist = filter(lambda x: x.selected, nlist)
        selection.start()
        selection.clear()
        selection.select(ilist)

registeredclass.Registration('Intersect Group',
                             NodeSelectionModifier, NodeIntersectGroup,
                             ordering=6,
                             params=[
    skeletongroupparams.NodeGroupParameter('group', tip="Name of the group.")],
                             tip='Select only the intersection of a group and the current selection.',
                             discussion="<para>Intersection between the specified <link linkend='Section:Concepts:Skeleton:Groups'>group</link> and the current selection. If there's none intersecting, none will be selected -- alternative way of <link linkend='MenuItem:OOF.NodeSelection.Clear'>clear</link>ing the selection.</para>")


###########################################################################

# Segments.

class SegmentSelectionModifier(registeredclass.RegisteredClass):
    registry = []
    def getSelection(self, skelcontext):
        return skelcontext.segmentselection

class SegFromSelectedElements(SegmentSelectionModifier):
    def __init__(self, internal=0, boundary=1):
        self.internal = internal
        self.boundary = boundary

    def getAllSegments(self, context):
        segments = {}
        skel = context.getObject()
        for element in context.elementselection.retrieve():
            for i in range(element.nnodes()):
                n0 = element.nodes[i]
                n1 = element.nodes[(i+1)%element.nnodes()]
                seg = skel.getSegment(n0, n1)
                segments[seg] = None
        return segments.keys()

    def getBoundarySegments(self, context):
        bound = []
        skel = context.getObject()
        for element in context.elementselection.retrieve():
            for i in range(element.nnodes()):
                n0 = element.nodes[i]
                n1 = element.nodes[(i+1)%element.nnodes()]
                seg = skel.getSegment(n0, n1)
                # A segment is on the boundary of the selection if it
                # belongs to only one element.
                n = 0
                for el in seg.getElements():
                    if el.selected: n += 1
                if n == 1:
                    bound.append(seg)
        return bound

    def getInternalSegments(self, context, allsegs):
        bound = self.getBoundarySegments(context)
        internal = []
        for seg in allsegs:
            if seg not in bound:
                internal.append(seg)
        return internal

    def __call__(self, skeleton, selection):
        if self.internal and self.boundary:
            selected = self.getAllSegments(skeleton)
        elif not self.internal and self.boundary:
            selected = self.getBoundarySegments(skeleton)
        elif self.internal and not self.boundary:
            allsegs = self.getAllSegments(skeleton)
            selected = self.getInternalSegments(skeleton, allsegs)
        else:
            selected = []
        selection.start()
        selection.clear()
        selection.select(selected)

registeredclass.Registration('Select From Selected Elements',
                             SegmentSelectionModifier,
                             SegFromSelectedElements, ordering=1,
                             params = [
    parameter.BooleanParameter('internal', 0, tip='Select internal segments.'),
    parameter.BooleanParameter('boundary', 1, tip='Select boundary segments.')],
                             tip="Select segments from selected elements.",
                             discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu-segments_from_elements.xml'))


class SelectInternalBoundarySegments(SegmentSelectionModifier):
    def __call__(self, skeleton, selection):
        skel = skeleton.getObject()
        seglist = []
        for segment in skel.segments.values():
            elements = segment.getElements()
            if len(elements) == 2 and elements[0].dominantPixel(skel) != \
                   elements[1].dominantPixel(skel):
                seglist.append(segment)
        selection.start()
        selection.clear()
        selection.select(seglist)

registeredclass.Registration('Select Internal Boundary Segments',
                             SegmentSelectionModifier,
                             SelectInternalBoundarySegments,
                             ordering=1.25,
                             tip="Select segments on material or group boundaries.",
                             discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu-boundary_segments.xml'))

class SegmentHomogeneity(SegmentSelectionModifier):
    def __init__(self, threshold=0.9):
        self.threshold = threshold

    def __call__(self, skeleton, selection):
        selected = []
        skel = skeleton.getObject()
        for segment in skel.segment_iterator():
            if segment.homogeneity(skel.MS) <= self.threshold:
                selected.append(segment)
        selection.start()
        selection.clear()
        selection.select(selected)

registeredclass.Registration('Select By Homogeneity', SegmentSelectionModifier,
                             SegmentHomogeneity, ordering=1.5,
                             params = [
    parameter.FloatRangeParameter('threshold', (0.0, 1.0, 0.01), value=0.9,
                              tip='The threshold homogeneity.')],
                             tip="Select segments with homogeneity less than the given threshold.",
                             discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu-hetero_segments.xml'))

# Select the indicated group.
class SegmentSelectGroup(SegmentSelectionModifier):
    def __init__(self, group):
        self.group = group
    def __call__(self, skeleton, selection):
        # Group retrieval may throw an exception -- do it first.
        members = skeleton.segmentgroups.get_group(self.group)
        selection.start()
        selection.clear()
        selection.select(members)

registeredclass.Registration('Select Group',
                             SegmentSelectionModifier, SegmentSelectGroup,
                             ordering=2,
                             params=[
    skeletongroupparams.SegmentGroupParameter('group', tip="Name of the group")],
                             tip='Select the members of a group, keeping the current selection.')



# Unselect the indicated group.
class SegmentDeselectGroup(SegmentSelectionModifier):
    def __init__(self, group):
        self.group = group
    def __call__(self, skeleton, selection):
        members = skeleton.segmentgroups.get_group(self.group)
        selection.start()
        selection.deselect(members)

registeredclass.Registration('Unselect Group',
                             SegmentSelectionModifier, SegmentDeselectGroup,
                             ordering=3,
                             params=[
    skeletongroupparams.SegmentGroupParameter('group', "Name of the group.")],
                             tip='Unselect the members of a group.')


# Add the group to the selection, retaining the current selection.
class SegmentAddSelectGroup(SegmentSelectionModifier):
    def __init__(self, group):
        self.group = group
    def __call__(self, skeleton, selection):
        members = skeleton.segmentgroups.get_group(self.group)
        selection.start()
        selection.select(members)

registeredclass.Registration('Add Group',
                             SegmentSelectionModifier, SegmentAddSelectGroup,
                             ordering=4,
                             params=[
    skeletongroupparams.SegmentGroupParameter('group', tip="Name of the group.")],
                             tip='Select the members of a group, retaining the current selection.')


# Select the intersection of the group and the selection.
class SegmentIntersectGroup(SegmentSelectionModifier):
    def __init__(self, group):
        self.group = group
    def __call__(self, skeleton, selection):
        slist = skeleton.segmentgroups.get_group(self.group)
        ilist = filter(lambda x: x.selected, slist)
        selection.start()
        selection.clear()
        selection.select(ilist)

registeredclass.Registration('Intersect Group',
                             SegmentSelectionModifier, SegmentIntersectGroup,
                             ordering=5,
                             params=[
    skeletongroupparams.SegmentGroupParameter('group', tip="Name of the group.")],
                             tip='Select only the intersection of a group and the current selection.')


###########################################################################

# Elements

class ElementSelectionModifier(registeredclass.RegisteredClass):
    registry = []
    def getSelection(self, skelcontext):
        return skelcontext.elementselection

class ByElementType(ElementSelectionModifier):
    def __init__(self, shape):
        self.shape = shape

    def __call__(self, skeleton, selection):
        selected = []
        for element in skeleton.getObject().element_iterator():
            if element.type() == self.shape:
                selected.append(element)
        selection.start()
        selection.clear()
        selection.select(selected)

registeredclass.Registration('Select By Element Type',
                             ElementSelectionModifier,
                             ByElementType, ordering=1,
                             params = [
    enum.EnumParameter('shape', skeletonelement.ElementShapeType,
                       skeletonelement.ElementShapeType('triangle'),
                       tip="Element shape.")],
                             tip='Select elements by element-shape.')

class ElementHomogeneity(ElementSelectionModifier):
    def __init__(self, threshold=0.9):
        self.threshold = threshold

    def __call__(self, skeleton, selection):
        selected = []
        skel = skeleton.getObject()
        for element in skel.element_iterator():
            if element.homogeneity(skel) < self.threshold:
                selected.append(element)
        selection.start()
        selection.clear()
        selection.select(selected)

registeredclass.Registration('Select By Homogeneity', ElementSelectionModifier,
                             ElementHomogeneity, ordering=2,
                             params = [
    parameter.FloatRangeParameter('threshold', (0.0, 1.0, 0.01), value=0.9,
                              tip='Threshold homogenity.')],
                             tip="Select Elements with homogeneity less than the threshold homogeneity.")

class ElementShapeEnergy(ElementSelectionModifier):
    def __init__(self, threshold = 0.8):
        self.threshold = threshold
    def __call__(self, skeleton, selection):
        selected = []
        for element in skeleton.getObject().element_iterator():
            if element.energyShape() > self.threshold:
                selected.append(element)
        selection.start()
        selection.clear()
        selection.select(selected)

registeredclass.Registration('Select By Shape Energy',
                             ElementSelectionModifier,
                             ElementShapeEnergy, ordering=3,
                             params = [
    parameter.FloatRangeParameter('threshold', (0.0, 1.0, 0.01), value=0.8,
                                  tip='Elements with shape-energy greater than the threshold will be selected.')],
                             tip="Select elements by shape-energy. The greater the shape-energy the uglier the element.")

class ElementIllegal(ElementSelectionModifier):
    def __call__(self, skeleton, selection):
        selected = []
        for element in skeleton.getObject().element_iterator():
            if element.illegal():
                selected.append(element)
        selection.start()
        selection.clear()
        selection.select(selected)

registeredclass.Registration('Select Illegal Elements',
                             ElementSelectionModifier,
                             ElementIllegal, ordering=3.05,
                             tip="Select illegal elements.")

class ElementFromSelectedNodes(ElementSelectionModifier):
    def __call__(self, skeleton, selection):
        selected = {}
        for node in skeleton.nodeselection.retrieve():
            for el in node.neighborElements():
                selected[el] = None
        selection.start()
        selection.clear()
        selection.select(selected.keys())

registeredclass.Registration('Select From Selected Nodes',
                             ElementSelectionModifier,
                             ElementFromSelectedNodes, ordering=3.1,
                             tip="Select every element containing any of selected nodes.")

class ElementFromSelectedSegments(ElementSelectionModifier):
    def __call__(self, skeleton, selection):
        selected = {}
        for segment in skeleton.segmentselection.retrieve():
            for el in segment.getElements():
                selected[el] = None
        selection.start()
        selection.clear()
        selection.select(selected.keys())

registeredclass.Registration('Select from Selected Segments',
                             ElementSelectionModifier,
                             ElementFromSelectedSegments, ordering=3.15,
                             tip="Select every element adjacent to selected segments.")

class ExpandElementSelection(ElementSelectionModifier):
    def boundary(self, el):
        for n in el.nodes:
            for e in n.neighborElements():
                if not e.selected:
                    return 1
        return 0

    def __call__(self, skeleton, selection):
        selected = {}
        # Boundary elements : not all of their neighbors are selected
        belements = []
        for el in selection.retrieve():
            if self.boundary(el):
                belements.append(el)
            else:
                selected[el] = None
        for be in belements:
            for n in be.nodes:
                for e in n.neighborElements():
                    selected[e] = None
            selected[be] = None
        selection.start()
        selection.clear()
        selection.select(selected.keys())
    
registeredclass.Registration('Expand Element Selection',
                             ElementSelectionModifier,
                             ExpandElementSelection, ordering=3.2,
                             tip="Expand element selection by nodes -- any elements that are connected to nodes of selected elements will be added to the selection.")

# Select the indicated group.
class ElementSelectGroup(ElementSelectionModifier):
    def __init__(self, group):
        self.group = group
    def __call__(self, skeleton, selection):
        members = skeleton.elementgroups.get_group(self.group)
        selection.start()
        selection.clear()
        selection.select(members)

registeredclass.Registration('Select Group',
                             ElementSelectionModifier, ElementSelectGroup,
                             ordering=4,
                             params=[
    skeletongroupparams.ElementGroupParameter('group', tip="Name of the group.")],
                             tip='Select the members of a group, keeping the current selection.')


# Unselect the indicated group.
class ElementDeselectGroup(ElementSelectionModifier):
    def __init__(self, group):
        self.group = group
    def __call__(self, skeleton, selection):
        members = skeleton.elementgroups.get_group(self.group)
        selection.start()
        selection.deselect(members)

registeredclass.Registration('Unselect Group',
                             ElementSelectionModifier, ElementDeselectGroup,
                             ordering=5,
                             params=[
    skeletongroupparams.ElementGroupParameter('group', tip="Name of the group.")],
                             tip='Unselect the members of a group.')


# Add the group to the selection, retaining the current selection.
class ElementAddSelectGroup(ElementSelectionModifier):
    def __init__(self, group):
        self.group = group
    def __call__(self, skeleton, selection):
        members = skeleton.elementgroups.get_group(self.group)
        selection.start()
        selection.select(members)


registeredclass.Registration('Add Group',
                             ElementSelectionModifier, ElementAddSelectGroup,
                             ordering=6,
                             params=[
    skeletongroupparams.ElementGroupParameter('group', tip="Name of the group.")],
                             tip='Select the members of a group, retaining the current selection.')


# Select the intersection of the group and the selection.
class ElementIntersectGroup(ElementSelectionModifier):
    def __init__(self, group):
        self.group = group
    def __call__(self, skeleton, selection):
        elist = skeleton.elementgroups.get_group(self.group)
        ilist = filter(lambda x: x.selected, elist)
        selection.start()
        selection.clear()
        selection.select(ilist)


registeredclass.Registration('Intersect Group',
                             ElementSelectionModifier, ElementIntersectGroup,
                             ordering=7,
                             params=[
    skeletongroupparams.ElementGroupParameter('group', tip="Name of the group.")],
                             tip='Select only the intersection of a group and the current selection.')
