// -*- C++ -*-
// $RCSfile: symmmatrix.h,v $
// $Revision: 1.18 $
// $Author: redwing $
// $Date: 2004/10/21 00:33:25 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>


// Symmetric matrix storage class
// not for linear algebra!

class SymmMatrix;

#ifndef SYMMMATRIX_H
#define SYMMMATRIX_H

#include <iostream>
#include "engine/eigenvalues.h"
#include "engine/matvec.h"
#include "engine/outputval.h"
#include "engine/fieldindex.h"

class IndexP;

class SymmMatrix {
protected:
//   friend double SymmMatrix_get(SymmMatrix*, int, int);
  double **m;
  int nrows;
  void allocate();
  void free();
  virtual void copy(double**);
public:
  SymmMatrix() : m(0), nrows(0) {}
  SymmMatrix(int);		// specifies size
  SymmMatrix(const SymmMatrix&); // copy constructor
  virtual ~SymmMatrix();
  SymmMatrix &operator=(const SymmMatrix&);
  SymmMatrix &operator*=(double);
  SymmMatrix &operator/=(double);
  double &operator()(int i, int j);
  double operator()(int i, int j) const;
  SymmMatrix &operator+=(const SymmMatrix&);
  SymmMatrix &operator-=(const SymmMatrix&);
  void resize(int);
  int size() const { return nrows; }
  void clear(double x=0);
  int badindex(int i) const { return i < 0 || i >= nrows; }
	
  SymmMatrix transform(const MATRIX_D&) const; // A^T B A
	
  friend class Debug;
  friend class SymmMatrix3;
  friend std::ostream& operator<<(std::ostream&, const SymmMatrix&);
  friend MATRIX_D operator*(const SymmMatrix&, const SymmMatrix&);
  friend SymmMatrix operator*(double, const SymmMatrix&);
  friend SymmMatrix operator*(const SymmMatrix&, double);
  friend VECTOR_D operator*(const SymmMatrix&, const VECTOR_D&);
};

class SymmMatrix3 : public OutputVal, public SymmMatrix {
// OutputVal is a PythonExportable class, and must be the first base
// class listed so that the PythonExportable dynamic classes work.
// This doesn't feel right...
private:
  mutable EigenValues eigenvalues; // cached
  mutable bool dirtyeigs_;	// are eigenvalues up-to-date?
  void findEigenvalues() const;
  static std::string classname_; // OutputVal is PythonExportable
  static std::string modulename_;
public:
  SymmMatrix3() : SymmMatrix(3), dirtyeigs_(true) {}
//   virtual ~SymmMatrix3();
  SymmMatrix3(double, double, double, double, double, double); // voigt order
  SymmMatrix3(const SymmMatrix3&);
  SymmMatrix3(const SymmMatrix&);
  virtual const std::string &classname() const { return classname_; }
  virtual const std::string &modulename() const { return modulename_; }
  SymmMatrix3 &operator=(const SymmMatrix3 &x) {
    dirtyeigs_ = x.dirtyeigs_;
    eigenvalues = x.eigenvalues;
    return dynamic_cast<SymmMatrix3&>(SymmMatrix::operator=(x));
  }

  // Vestigial __getitem__ -- takes a SymTensorIndex, but in practice,
  // values are actually retrieved via "get" with an (i,j) pair.
  // double __getitem__(const SymTensorIndex* sti) {
  //   return this->operator()(sti->row(),sti->col());
  // }


//   SymmMatrix3 &operator*=(double x) {
//     dirtyeigs_ = true;
//     return dynamic_cast<SymmMatrix3&>(SymmMatrix::operator*=(x));
//   }
  OutputVal &operator*=(double x) {
    dirtyeigs_ = true;
    return dynamic_cast<OutputVal&>(SymmMatrix::operator*=(x));
  }
  OutputVal &operator+=(const OutputVal &x) {
    dirtyeigs_ = true;
    const SymmMatrix3 &sm = dynamic_cast<const SymmMatrix3&>(x);
    return dynamic_cast<OutputVal&>(SymmMatrix::operator+=(sm));
  }
  OutputVal &operator-=(const OutputVal &x) {
    dirtyeigs_ = true;
    const SymmMatrix3 &sm = dynamic_cast<const SymmMatrix3&>(x);
    return dynamic_cast<OutputVal&>(SymmMatrix::operator-=(sm));
  }
  SymmMatrix3 &operator/=(double x) {
    dirtyeigs_ = true;
    return dynamic_cast<SymmMatrix3&>(SymmMatrix::operator/=(x));
  }
  SymmMatrix3 &operator+=(const SymmMatrix3 &x) {
    dirtyeigs_ = true;
    return dynamic_cast<SymmMatrix3&>(SymmMatrix::operator+=(x));
  }
  SymmMatrix3 &operator-=(const SymmMatrix3 &x) {
    dirtyeigs_ = true;
    return dynamic_cast<SymmMatrix3&>(SymmMatrix::operator-=(x));
  }
  virtual double operator[](const IndexP&) const;
  virtual double &operator[](const IndexP&);
  double trace() const;
  double determinant() const;
  double secondInvariant() const;
  double maxEigenvalue() const;
  double midEigenvalue() const;
  double minEigenvalue() const;

  virtual IndexP getIndex(const std::string&) const;
  virtual IteratorP getIterator() const;
  virtual void print(std::ostream&) const;
};

OutputValue *newSymTensorOutputValue();

//ostream& operator<<(ostream&, const SymmMatrix&);



#endif
