# -*- python -*-
# $RCSfile: imagemodifier.py,v $
# $Revision: 1.26 $
# $Author: haan $
# $Date: 2004/11/17 16:39:08 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from SWIG.common import switchboard
from SWIG.image import oofimage
from common import debug
from common import enum
from common import registeredclass
from common.IO import parameter
from SWIG.image.GRAINBDY import imageops
from types import *

# Base class for Python ImageModifiers. 
class ImageModifier(registeredclass.RegisteredClass):
    registry = []
    def __call__(self, image):
        pass

# OOFMenu callback, installed automatically for each ImageModifier
# class by the switchboard callback invoked when the class is
# registered.

def doImageMod(menuitem, image, **params):
    # image is the image name, actually
    imagecontext = oofimage.imageContexts[image]
    imagecontext.reserve()
    try:
        immidge = imagecontext.getObject()  # OOFImage object
        newimmidge = immidge.clone("")
        registration = menuitem.data
        imageModifier = registration(**params) # create ImageModifier obj
        imagecontext.begin_writing()
        try:
            imageModifier(newimmidge)   # call its __call__ method on the image
            oofimage.pushModification(image, newimmidge)
        finally:
            imagecontext.end_writing()
    finally:
        imagecontext.cancel_reservation()
    switchboard.notify('modified image', imageModifier, image)
    switchboard.notify('redraw')

###################################

class FlipDirection(enum.EnumClass(
    ('x', 'Flip the image about the x axis'),
    ('y', 'Flip the image about the y axis'),
    ('xy', 'Flip the image about both the x and y axes (ie, rotate by 180 degrees)'))):
    pass

FlipDirection.tip = "Direction of image flip. See <xref linkend='MenuItem:OOF.Image.Modify.Flip'/>."
FlipDirection.discussion = "<para>TODO: Add discussion.</para>"

# Actual ImageModifier classes are derived from ImageModifier like this:

class FlipImage(ImageModifier):
    def __init__(self, axis):           # constructor
        self.axis = axis                # 'x' or 'y'
    def __call__(self, image):          # called by doImageMod
        image.flip(self.axis.name)


# Registering the FlipImage class like this installs it in the menus
# and GUI.  The names of the Parameters in the params list *must* be
# the same as the arguments to the __init__ method.

registeredclass.Registration('Flip',    # name appearing in menus & GUI
                             ImageModifier, # base class
                             FlipImage, # derived class
                             ordering=1.0, # position in menus
                             params=[   # list of constructor arguments
    enum.EnumParameter('axis',         # argument name
                       FlipDirection,     # argument type
                       FlipDirection('x'), # initial value
                       tip="Flip the image about this axis") # helpful hint
    ],
                             tip="Flip the image about the x or y axis."
                             )

#####################################

class GrayImage(ImageModifier):
    def __call__(self, image):
        image.gray()
registeredclass.Registration('Gray', ImageModifier, GrayImage,
                             ordering=0.5,
                             tip='Convert image to gray scale.')

class FadeImage(ImageModifier):
    def __init__(self, factor):
        self.factor = factor
    def __call__(self, image):
        image.fade(self.factor)

registeredclass.Registration('Fade', ImageModifier, FadeImage, ordering=1.1,
                             params=[
    parameter.FloatRangeParameter('factor', (0, 1, 0.01), 0.1,
                                  tip="0 (false) does nothing, 1 (true) fades to white.")],
                             tip="Fade the image by the given factor.")

class DimImage(ImageModifier):
    def __init__(self, factor):
        self.factor = factor
    def __call__(self, image):
        image.dim(self.factor)

registeredclass.Registration('Dim', ImageModifier, DimImage, ordering=1.2,
                             params=[
    parameter.FloatRangeParameter('factor', (0, 1, 0.01), value=0.9,
                              tip="0 (false) fades to black, 1 (true) does nothing.")
    ],
                             tip="Dim the image by the given factor.")

class BlurImage(ImageModifier):
    def __init__(self, radius, sigma):
        self.radius = radius
        self.sigma = sigma
    def __call__(self, image):
        image.blur(self.radius, self.sigma)

registeredclass.Registration('Blur', ImageModifier, BlurImage,
                             ordering=2.00,
                             params=[
    parameter.FloatParameter('radius', 0.0,
      tip="Radius of the Gaussian, in pixels, not counting the center pixel."),
    parameter.FloatParameter('sigma', 1.0,
      tip="Standard deviation of the Gaussian, in pixels")
    ],
                             tip=" Blurs an image.  Convolves the image with a Gaussian operator of the given radius and standard deviation (sigma). For reasonable results, the radius should be larger than sigma.  Use a radius of 0 and BlurImage() selects a suitable radius for you. [ImageMagick]")

class ContrastImage(ImageModifier):
    def __init__(self, sharpen):
        self.sharpen = sharpen
    def __call__(self, image):
        image.contrast(self.sharpen)

registeredclass.Registration('Contrast', ImageModifier, ContrastImage,
                             ordering=2.02,
                             params=[
    parameter.BooleanParameter('sharpen', 1, tip='0 (false) to dull, 1 (true) to sharpen')
    ],
                             tip="Enhance the intensity differences between the lighter and darker elements of the image.  Set sharpen to 'true' to increase the image contrast, otherwise the contrast is reduced. [ImageMagick]")
                             
class DespeckleImage(ImageModifier):
    def __call__(self, image):
        image.despeckle()

registeredclass.Registration('Despeckle', ImageModifier, DespeckleImage,
                             ordering=2.03,
                             tip= "Despeckle reduces the speckle noise in an image while preserving the edges of the original image. [ImageMagick]")


class EdgeImage(ImageModifier):
    def __init__(self, radius):
        self.radius = radius
    def __call__(self, image):
        image.edge(self.radius)
registeredclass.Registration('Edge', ImageModifier, EdgeImage,
                             ordering=2.031,
                             params=[
    parameter.FloatParameter('radius', 0.0, tip="Radius for the operation.")
    ],
                             tip="Find edges in an image.  Radius defines the radius of the convolution filter.  Use a radius of 0 and Edge selects a suitable radius for you. [ImageMagick]")

class EnhanceImage(ImageModifier):
    def __call__(self, image):
        image.enhance()

registeredclass.Registration('Enhance', ImageModifier, EnhanceImage,
                             ordering=2.04,
                             tip='Enhance (minimize noise), from ImageMagick')

class EqualizeImage(ImageModifier):
    def __call__(self, image):
        image.equalize()
registeredclass.Registration('Equalize', ImageModifier, EqualizeImage,
                             ordering=2.05,
                             tip='Histogram equalization [ImageMagick]')

class MedianFilterImage(ImageModifier):
    def __init__(self, radius):
        self.radius = radius
    def __call__(self, image):
        image.medianFilter(self.radius)
registeredclass.Registration('MedianFilter', ImageModifier, MedianFilterImage,
                             ordering=2.06,
                             params=[
    parameter.FloatParameter('radius', 1.0, tip="Radius for the filter.")],
                             tip="Replace each pixel component with the median color in a circular neighborhood.")

class NegateImage(ImageModifier):
    def __call__(self, image):
        image.negate(0)                 # 0 => negate all colors, not just gray
registeredclass.Registration('Negate', ImageModifier, NegateImage,
                             ordering=2.065,
                             tip="Negate negates the colors in the image. [ImageMagick]")

class NormalizeImage(ImageModifier):
    def __call__(self, image):
        image.normalize()
registeredclass.Registration('Normalize', ImageModifier, NormalizeImage,
                             ordering=2.07,
                             tip="Normalize image (increase contrast by normalizing the pixel values to span the full range of color values). [ImageMagick]"
                             )

class ReduceNoise(ImageModifier):
    def __init__(self, radius):
        self.radius = radius
    def __call__(self, image):
        image.reduceNoise(self.radius)
registeredclass.Registration('ReduceNoise', ImageModifier, ReduceNoise,
                             ordering=2.08,
                             params=[
    parameter.FloatParameter('radius', 0.0, tip='Size of the pixel neighborhood.')],
                             tip="Smooth the contours of an image while still preserving edge information.  The algorithm works by replacing each pixel with its neighbor closest in value.  A neighbor is defined by the radius.  Use a radius of 0 and ReduceNoise selects a suitable radius for you.  [ImageMagick]")

class SharpenImage(ImageModifier):
    def __init__(self, radius, sigma):
        self.radius = radius
        self.sigma = sigma
    def __call__(self, image):
        image.sharpen(self.radius, self.sigma)
registeredclass.Registration('Sharpen', ImageModifier, SharpenImage,
                             ordering=2.09,
                             params=[
    parameter.FloatParameter('radius', 0.0, tip='Radius of the Gaussian.'),
    parameter.FloatParameter('sigma', 1.0, tip='Standard deviation of the Gaussian.')
    ],
                             tip="Sharpen an image by convolving the image with a Gaussian operator of the given radius and standard deviation (sigma). For reasonable results, radius should be larger than sigma.  Use a radius of 0 and Sharpen selects a suitable radius for you. [ImageMagick]")

class ReIlluminateImage(ImageModifier):
    def __init__(self, radius):
        self.radius = radius
    def __call__(self, image):
        image.evenly_illuminate(self.radius)
registeredclass.Registration('Reilluminate', ImageModifier, ReIlluminateImage,
                             ordering=3.0,
                             params=[
    parameter.IntParameter('radius', 10, tip='Size of the averaging region.')
    ],
                             tip='Adjust brightness so that the whole image is evenly illuminated.')

class ThresholdImage(ImageModifier):
    def __init__(self, T):
        self.T=T
    def __call__(self, image):
        dbls = imageops.grayify(image)
        bools = oofimage.threshold(dbls,self.T)
        imageops.setFromBool(image,bools)
        
registeredclass.Registration(
    'ThresholdImage',
    ImageModifier,
    ThresholdImage,
    ordering=100,
    params=[
    parameter.FloatRangeParameter('T', (0,1,.01),value=.5,
                             tip="Threshold value between 0 and 1."),
    ],
    tip="Threshold an image given a threshold value."
)
